from marshmallow import Schema, fields


class JobSchema(Schema):
    task_id = fields.Str()
    uuid = fields.Str()
    status = fields.Str()
    result_path = fields.Str()
    batch_id = fields.Str()


class BatchSchema(Schema):
    batch_id = fields.Str()
    progress = fields.Float()
    result_url = fields.Str()
    finished = fields.Boolean()
    jobs_done = fields.List(fields.Nested(JobSchema))
    jobs_pending = fields.List(fields.Nested(JobSchema))



import pickle
import json
from abc import ABC, abstractclassmethod, abstractmethod
from typing import Callable, Dict, Iterable, Optional, Type
from uuid import uuid1

from flask import make_response, Response, request, url_for
from flask_apispec import marshal_with, use_kwargs
from flask_apispec.views import MethodResource
from marshmallow import fields
import requests

from .schema import JobSchema, BatchSchema


class ModelABC(object):
    @abstractclassmethod
    def create(cls, **kwargs) -> Response:
        raise NotImplementedError

    @abstractmethod
    def save(self):
        raise NotImplementedError

    def update(self, **kwargs):
        for key, value in kwargs.items():
            self.__setattr__(key, value)

        self.save()
        return self

    @abstractclassmethod
    def get(cls) -> Iterable:
        raise NotImplementedError

    @abstractclassmethod
    def get_by_id(cls, uuid: str):
        raise NotImplementedError

    @abstractclassmethod
    def delete(cls, uuid: str):
        raise NotImplementedError


Model = Type[ModelABC]


class TaskLoggerABC(object):
    @abstractclassmethod
    def save_job(cls, job_data):
        raise NotImplementedError


TaskLogger = Type[TaskLoggerABC]


class TaskManagerABC(object):
    def _create_task(self, uuid, binary_data, handler_uri):
        raise NotImplementedError

    def _enqueue(self, task):
        raise NotImplementedError


TaskManager = Type[TaskManagerABC]


class ResourceABC(MethodResource):
    _default_serializer: Callable = pickle.dumps
    _model: Model = ModelABC
    _task_manager: TaskManager = TaskManagerABC
    _task_router: Dict = dict()

    def _get(self, uuid):
        return self._model.get_by_id(uuid)

    def _create_task(self, binary_data: bytes, handler_uri: str):
        return self._task_manager._create_task(binary_data, handler_uri)

    def _enqueue(self, task: Dict):
        return self._task_manager._enqueue(task)

    @classmethod
    def register_to_blueprint(cls, blueprint):
        return blueprint.add_url_rule(cls.url, view_func=cls.as_view(cls.endpoint_name))


class JobListResourceABC(ResourceABC):
    url = "/"
    endpoint_name = "joblist_resource_api"

    @marshal_with(JobSchema(many=True))
    def get(self):
        return self._model.get()

    @marshal_with(JobSchema)
    @use_kwargs(
        {
            "uuid": fields.Str(),
            "task_type": fields.Str(),
            "json_data": fields.Str(),
            "batch_id": fields.Str(),
        }
    )
    def post(self, task_type, json_data, batch_id=None, uuid=None):
        uuid = uuid or uuid1().hex
        serializer = self._task_router.get("serializer") or self._default_serializer
        data = json.loads(json_data)
        data["uuid"] = uuid
        binary_data = serializer(data)
        task = self._create_task(binary_data, self._task_router.get(task_type))
        response = self._enqueue(task)
        job_data = dict(
            task_id=response.name,
            task_type=task_type,
            status="waiting",
            result_path=None,
            uuid=uuid,
            batch_id=batch_id,
        )
        job = self._model.create(**job_data)

        return job, 201


class JobResourceABC(ResourceABC):
    url = "/<uuid>/"
    endpoint_name = "job_resource_api"

    @marshal_with(JobSchema)
    def get(self, uuid):
        return self._get(uuid)

    @marshal_with(JobSchema)
    @use_kwargs(JobSchema)
    def put(self, uuid, **kwargs):
        job = self._get(uuid)
        if job is None:
            self._model.create(uuid=uuid, **kwargs)

        else:
            job.update(**kwargs)

        return job, 201

    def delete(self, uuid):
        job = self._get(uuid)
        job.delete()
        return {}, 204


class BatchListResourceABC(ResourceABC):
    url = "/batches/"
    endpoint_name = "batchlist_resource_api"

    @marshal_with(BatchSchema(many=True))
    def get(self):
        return self._model.get()

    @marshal_with(BatchSchema)
    @use_kwargs(
        {
            "batch_id": fields.Str(),
            "task_types": fields.List(fields.Str()),
            "json_datas": fields.List(fields.Str()),
            "task_ids": fields.List(fields.Str()),
        }
    )
    def post(self, task_types, json_datas, batch_id=None, task_ids=None):
        batch_uuid = batch_id or uuid1().hex
        task_ids = task_ids or [uuid1().hex] * len(task_types)

        for task_id, task_type, json_data in zip(task_ids, task_types, json_datas):
            r = requests.post(
                request.url_root + url_for("QueueManager.joblist_resource_api"),
                data=dict(
                    uuid=task_id,
                    task_type=task_type,
                    json_data=json_data,
                    batch_id=batch_uuid,
                ),
            )

        batch = self._model.create(batch_id=batch_uuid)
        return batch, 201


class BatchResourceABC(ResourceABC):
    url = "/batches/<uuid>/"
    endpoint_name = "batch_resource_api"

    @marshal_with(BatchSchema)
    def get(self, uuid):
        return self._model.get_and_format_by_id(uuid)

    @marshal_with(BatchSchema)
    @use_kwargs(BatchSchema)
    def put(self, uuid, **kwargs):
        batch = self._get(uuid)
        if batch is None:
            self._model.create(uuid=uuid, **kwargs)

        else:
            batch.update(**kwargs)

        return batch, 201




import json

import requests

url = "http://localhost:8080/queue/batches/"
task_types = ["say-hello"] * 10
json_datas = [json.dumps({"hello": "world"})] * 10
task_ids = [f"hello-{i}" for i in range(10)]


payload = dict(
    batch_id="batch-test",
    task_types=task_types,
    json_datas=json_datas,
    task_ids=task_ids,
)
r = requests.post(url, data=payload)



from dotenv import load_dotenv, find_dotenv

print(load_dotenv(find_dotenv()))

from flask import Flask

from kws_pipeline.google_helpers.task_manager import GoogleTaskManager
from kws_pipeline.google_helpers.job_model import Job, Batch
from kws_pipeline.google_helpers.middlewares import NDBClientMiddleware
from kws_pipeline.queue.utils import create_app

task_manager = GoogleTaskManager(
    project_id="hello-infra-kws",
    location="europe-west1",
    queue="simulations",
    worker_url="http://localhost:8081",
)

task_router = {
    "say-hello": "/worker/say-hello",
    "upper-greetings": "/worker/upper-greetings",
}

if __name__ == "__main__":
    queue_manager = create_app(
        job_model=Job,
        batch_model=Batch,
        task_manager=task_manager,
        task_router=task_router,
        middlewares=(NDBClientMiddleware,),
    )
    queue_manager.run(port=8080, debug=True)



from flask import Flask
from kws_pipeline.workers.worker import Worker
from kws_pipeline.workers.queue_manager import QueueManager
from kws_pipeline.google_helpers.bucket_manager import BucketStorage

queue_manager = QueueManager("http://localhost:8080/")
storage_manager = BucketStorage("test-bucket-hello-kws")


worker = Worker(queue_manager=queue_manager, storage_manager=storage_manager)


@worker.capability
def say_hello(greetings: str, to: str = "world"):
    """
    Says hello to someone
    """
    return f"{greetings}, {to}"


worker.pipeline("Upper greetings", say_hello, str.upper)
worker.build_router()

app = Flask(__name__)
worker.register(app)

if __name__ == "__main__":
    app.run(port=8081)



import pickle
import inspect
from urllib.parse import urljoin
from functools import wraps

import requests
from flask import Blueprint, request, jsonify
from slugify import slugify

from .queue_manager import QueueManager
from .storage_manager import StorageManagerABC


class Worker(object):
    def __init__(
        self, worker_prefix="worker", queue_manager=None, storage_manager=None
    ):
        self.worker_prefix = worker_prefix
        self.storage_manager: StorageManagerABC = storage_manager
        self.queue_manager: QueueManager = queue_manager
        self.blueprint = Blueprint("worker", __name__, url_prefix=f"/{worker_prefix}/")
        self.capabilities = set()

    def register_slug(self, name):
        slug = slugify(name)
        self.capabilities.add(slug)
        return f"/{slug}/"

    def build_documentation(self, func):
        sig = inspect.signature(func)
        doc = {"docstring": func.__doc__, "parameters": {}}
        params = doc["parameters"]
        for name, param in sig.parameters.items():
            param_doc = {}
            if param.annotation is not param.empty:
                param_doc["type"] = str(param.annotation)
            if param.default is not param.empty:
                param_doc["default"] = param.default
            params[name] = param_doc

        return doc

    def init_task(self, status="pending"):
        data = request.get_data()
        payload = pickle.loads(request.get_data())
        uuid = payload.pop("uuid")
        resp = self.queue_manager.set_status(uuid, status=status)
        return uuid, payload

    def post_task(self, uuid, result, status="done"):
        result_path = self.storage_manager.write_to_file(uuid, result)
        resp = self.queue_manager.set_status(
            uuid, result_path=result_path, status=status
        )
        self.queue_manager._reduce(uuid, self.storage_manager)

        return result_path

    def capability(self, func):
        slug = slugify(func.__name__)
        route = self.register_slug(func.__name__)

        def wrapped_function():
            uuid, payload = self.init_task()
            print(uuid, payload)
            result = func(**payload)
            print(result)
            result_path = self.post_task(uuid, result)
            return jsonify({"result_path": result_path}), 200

        def wrapped_doc():
            doc = self.build_documentation(func)
            return jsonify(doc), 200

        wrapped_function.__name__ = f"{slug}_worker"
        wrapped_doc.__name__ = f"{slug}_doc"
        self.blueprint.route(route, methods=("POST",))(wrapped_function)
        self.blueprint.route(route, methods=("GET",))(wrapped_doc)
        return func

    def pipeline(self, pipeline_name, *funcs):
        slug = slugify(pipeline_name)
        route = self.register_slug(pipeline_name)

        def unrolled(*args, **kwargs):
            func = funcs[0]
            result = func(*args, **kwargs)
            for func in funcs[1:]:
                result = func(result)

            return result

        def wrapped_pipeline():
            pending_template = "pending: %s"
            func = funcs[0]
            uuid, payload = self.init_task(status=pending_template % func.__name__)
            result = func(**payload)
            for func in funcs[1:]:
                self.queue_manager.set_status(
                    uuid, status=pending_template % func.__name__
                )
                result = func(result)

            result_path = self.post_task(uuid, result)
            return jsonify({"result_path": result_path}), 200

        def wrapped_doc():
            docs = [self.build_documentation(func) for func in funcs]
            return jsonify(docs), 200

        wrapped_pipeline.__name__ = f"{slug}-worker"
        wrapped_doc.__name__ = f"{slug}-doc"

        self.blueprint.route(route, methods=("POST",))(wrapped_pipeline)
        self.blueprint.route(route, methods=("GET",))(wrapped_doc)

        return unrolled

    def register(self, app):
        app.register_blueprint(self.blueprint)

    def build_router(self):
        router = dict()
        for capability in self.capabilities:
            router[capability] = f"/{self.worker_prefix}/{capability}/"

        @self.blueprint.route("/router/")
        def _router():
            nonlocal router
            return jsonify(router), 200

        return router



from google.cloud import ndb
from flask import make_response

from ..queue.abc import ModelABC


class Job(ModelABC, ndb.Model):
    task_id = ndb.StringProperty()
    batch_id = ndb.StringProperty()
    task_type = ndb.StringProperty()
    uuid = ndb.StringProperty()
    status = ndb.StringProperty()
    result_path = ndb.StringProperty()

    def save(self):
        super().put()

    @classmethod
    def create(cls, **kwargs):
        job = cls(**kwargs)
        job.put()
        return job

    @classmethod
    def get(cls):
        r = make_response({"name": "coucou"}, 201)
        return [j for j in cls.query()]

    @classmethod
    def get_by_id(cls, uuid):
        return cls.query().filter(cls.uuid == uuid).get()


class Batch(ModelABC, ndb.Model):
    batch_id = ndb.StringProperty()
    result_url = ndb.StringProperty()

    def save(self):
        super().put()

    @classmethod
    def get(cls):
        print("hello")
        return [b for b in cls.query()]

    @classmethod
    def create(cls, **kwargs):
        batch = cls(**kwargs)
        batch.put()
        return batch

    @classmethod
    def get_by_id(cls, batch_id):
        batch = cls.query().filter(cls.batch_id == batch_id).get()
        return batch

    @classmethod
    def get_and_format_by_id(cls, batch_id):
        batch = cls.query().filter(cls.batch_id == batch_id).get()
        jobs = Job.query().filter(Job.batch_id == batch_id)
        done = [j for j in jobs.filter(Job.status == "done")]
        pending = [j for j in jobs.filter(Job.status != "done")]
        finished = len(pending) == 0

        return {
            "batch_id": batch_id,
            "result_url": batch.result_url,
            "finished": finished,
            "jobs_done": done,
            "jobs_pending": pending,
            "progress": len(done) / (len(pending) + len(done)),
        }




import requests
import pickle

url = "http://localhost:8081/worker/%s"
data = dict(uuid="abc1", greetings="hello", to="world")

r = requests.post(url % "say-hello", data=pickle.dumps(data))
r = requests.post(url % "upper-greetings", data=pickle.dumps(data))
r = requests.post(url % "dictionary", data=pickle.dumps(dict(uuid="abc2")))



from google.cloud import ndb


class NDBClientMiddleware:
    client = ndb.Client()

    def __init__(self, wsgi_app):
        self.wsgi_app = wsgi_app

    def __call__(self, environ, start_response):
        with self.client.context():
            return self.wsgi_app(environ, start_response)



from flask import Flask, Blueprint

from kws_pipeline.queue.abc import (
    JobResourceABC,
    JobListResourceABC,
    BatchListResourceABC,
    BatchResourceABC,
)


def register_resources(app, model, batch_model, task_manager, task_router):
    class JobListResource(JobListResourceABC):
        _model = model
        _task_manager = task_manager
        _task_router = task_router

    class JobResource(JobResourceABC):
        _model = model
        _task_manager = task_manager
        _task_router = task_router

    class BatchListResource(BatchListResourceABC):
        _model = batch_model
        _task_manager = task_manager
        _task_router = task_router

    class BatchResource(BatchResourceABC):
        _model = batch_model
        _task_manager = task_manager
        _task_router = task_router

    JobListResource.register_to_blueprint(app)
    JobResource.register_to_blueprint(app)
    BatchListResource.register_to_blueprint(app)
    BatchResource.register_to_blueprint(app)
    return app


def apply_middlewares(app, *middlewares):
    for middleware in middlewares:
        app.wsgi_app = middleware(app.wsgi_app)

    return app


def create_app(job_model, batch_model, task_manager, task_router, middlewares=tuple(), url_prefix="/queue/"):
    app = Flask(__name__)
    blueprint = Blueprint("QueueManager", __name__, url_prefix=url_prefix)
    blueprint = register_resources(
        blueprint, job_model, batch_model, task_manager, task_router
    )
    app.register_blueprint(blueprint)
    app = apply_middlewares(app, *middlewares)
    return app



from urllib.parse import urljoin

import requests


class QueueManager:
    def __init__(self, queue_manager_url, reduction_method=None):
        self.queue_manager_url = queue_manager_url
        if reduction_method:
            self.reduction = reduction_method
        else:

            def _reduce(_, __):
                pass

            self.reduction = _reduce

    def set_status(self, uuid, **kwargs):
        url = urljoin(self.queue_manager_url, str(uuid))
        r = requests.put(url, data=kwargs)
        return r

    def _reduce(self, uuid, storage_manager):
        batch_id = self.get_batch_id(uuid)
        if batch_id:
            return self.reduce_batch(batch_id, storage_manager)
        return

    def reduce_batch(self, batch_id, storage_manager):
        batch = self.get_batch(batch_id)

        if batch["progress"] < 1:
            return

        data = {
            job["uuid"]: storage_manager.read_from_file(job["uuid"])
            for job in batch["jobs_done"]
        }

        batch_result = self.reduction(data, storage_manager, batch_id)

        url = urljoin(self.queue_manager_url, f"batches/{str(batch_id)}/")
        requests.put(url, {"result_url": batch_result})
        return batch_result

    def get_batch_id(self, uuid):
        return (
            requests.get(urljoin(self.queue_manager_url, str(uuid)))
            .json()
            .get("batch_id", None)
        )

    def get_batch(self, batch_id):
        return requests.get(
            urljoin(self.queue_manager_url, f"batches/{str(batch_id)}")
        ).json()



import json

import requests

url = "http://localhost:8080/queue/"
data = dict(num_rows=1000, headers=list("abcdefghijklmnopqrstuvwxyz"))
payload = dict(task_type="write-pandas-df", json_data=json.dumps(data))
r = requests.post(url, data=payload)



import pickle
from urllib.parse import urljoin

from ..queue.abc import TaskManagerABC

from google.cloud import tasks_v2


class GoogleTaskManager(TaskManagerABC):
    def __init__(self, project_id, location, queue, worker_url=""):
        self.task_client = tasks_v2.CloudTasksClient()
        self.parent = self.task_client.queue_path(project_id, location, queue)
        self.worker_url = worker_url

    def _create_task(self, binary_data, handler_uri):
        if not self.worker_url:
            task = {
                "app_engine_http_request": {
                    "http_method": "POST",
                    "relative_uri": handler_uri,
                    "body": binary_data,
                }
            }

        else:
            task = {
                "http_request": {
                    "url": urljoin(self.worker_url, handler_uri),
                    "http_method": "POST",
                    "body": binary_data,
                }
            }

        return task

    def _enqueue(self, task):
        return self.task_client.create_task(self.parent, task)



from ..workers.storage_manager import StorageManagerABC
from google.cloud import storage


class BucketStorage(StorageManagerABC):
    def __init__(self, bucket_name):
        self.client = storage.Client()
        self.bucket_name = bucket_name
        self.bucket = self.client.bucket(bucket_name)

    def write_to_file(self, uuid, result, content_type="text/plain"):
        blob = self.bucket.blob(uuid)
        blob.upload_from_string(result, content_type=content_type)
        return f"https://storage.cloud.google.com/{self.bucket_name}/{blob.name}"

    def read_from_file(self, uuid):
        blob = self.bucket.blob(uuid)
        return blob.download_as_string()
